"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-28 22:57:36
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-07 18:53:01
FilePath     : /wynntilsresolver/blocks/type.py
"""

from typing import List, Type

from wynntilsresolver.exception import InvalidItemType

from .block import Block
from .identification import Identifications
from .name import Name
from .powder import Powder
from .reroll import Reroll
from .shiny import Shiny


class ItemType(Block):
    _start_byte: int = 1
    _type_id: int
    _required_blocks: List[Type[Block]] = []
    _optional_blocks: List[Type[Block]] = []

    def to_bytes(self) -> List[int]:
        return self.encode_with_start([self._type_id])

    def __str__(self) -> str:
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class GearItem(ItemType):
    _type_id = 0
    _required_blocks = [Name]
    _optional_blocks = [Identifications, Powder, Reroll, Shiny]

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs) -> "GearItem":
        super().from_bytes(data)
        if data[0] != cls._type_id:
            raise InvalidItemType(f"Invalid item type {data[0]} for item {cls.__name__}")
        del data[0]
        return cls()



"""
Author       : FYWinds i@windis.cn
Date         : 2024-03-08 15:24:12
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-08 15:53:44
FilePath     : /wynntilsresolver/item.py
"""

# This file contains several pre-built resolver for specific items.

from __future__ import annotations

from typing import List, Optional

from .blocks import GearItem, Identification, Identifications, Name, Powder, Reroll, Shiny
from .resolver import Resolver


class GearItemResolver(Resolver):
    item_type: GearItem
    _name: Name
    _identifications: Optional[Identifications]
    powder: Optional[Powder]
    shiny: Optional[Shiny]
    _reroll: Optional[Reroll]

    @classmethod
    def create(
        cls,
        name: Name,
        identifications: Optional[Identifications] = None,
        powder: Optional[Powder] = None,
        shiny: Optional[Shiny] = None,
        reroll: Optional[Reroll] = None,
    ) -> "GearItemResolver":
        c = cls()
        c._name = name
        c._identifications = identifications
        c.powder = powder
        c.shiny = shiny
        c._reroll = reroll
        return c

    @property
    def name(self) -> str:
        return self._name.name

    @property
    def identifications(self) -> Optional[List[Identification]]:
        if self._identifications:
            return self._identifications.identifications

    @property
    def reroll(self) -> int:
        # According to current Artemis Algorithm, None means 0 rerolls
        if self._reroll:
            return self._reroll.rerolls
        return 0


__all__ = ["GearItemResolver"]



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-29 18:42:50
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-01 20:49:32
FilePath     : /src/wynntilsresolver/blocks/powder.py
"""

import math
from typing import List

from .block import Block

_POWDER_ELEMENTS = ["E", "T", "W", "F", "A"]


class Powder(Block):
    _start_byte = 4
    powder_slots: int
    powders: List[str]

    def __init__(self, powder_slots: int, powders: List[str]) -> None:
        self.powder_slots = powder_slots
        self.powders = powders

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs) -> "Powder":
        super().from_bytes(data)
        powder_slots = data[0]
        del data[0]
        powder_num = data[0]
        del data[0]

        # change the bytes to binary and pad it to 8 bits
        powder_binary = "".join([format(powder, "b").zfill(8) for powder in data[: math.ceil(powder_num * 5 / 8)]])
        del data[: math.ceil(powder_num * 5 / 8)]

        # 5 bit per powder
        powders_raw = [powder_binary[i : i + 5] for i in range(0, len(powder_binary), 5)]

        # convert to int
        powders_bin = [int(powder, 2) - 1 for powder in powders_raw if int(powder) != 0]
        powders = [f"{_POWDER_ELEMENTS[powder // 6]}{powder % 6 + 1}" for powder in powders_bin]

        return cls(powder_slots, powders)

    def to_bytes(self) -> List[int]:
        raise NotImplementedError

    def __str__(self) -> str:
        return f"{self.powders}, slots: {len(self.powders)}/{self.powder_slots}"

    def __repr__(self) -> str:
        return f"Powder({self.powders}, slots: {len(self.powders)}/{self.powder_slots})"



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-29 15:48:21
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-06-02 11:34:26
FilePath     : /wynntilsresolver/startup.py
"""

import asyncio
import json
import os
import pathlib
import time

import httpx

from .utils import run_async

if os.name == "nt":
    DATA_LOCATION = pathlib.Path.home() / "AppData" / "Local" / "wynntilsresolver"
else:
    DATA_LOCATION = pathlib.Path.home() / ".local" / "share" / "wynntilsresolver"

if data_location := os.environ.get("DATA_LOCATION"):
    DATA_LOCATION = pathlib.Path(data_location)

DATA_LOCATION.mkdir(parents=True, exist_ok=True)


ITEMDB_FULL_URL = "https://api.wynncraft.com/v3/item/database?fullResult"
SHINY_TABLE_URL = "https://raw.githubusercontent.com/Wynntils/Static-Storage/main/Data-Storage/shiny_stats.json"
ID_TABLE_URL = "https://raw.githubusercontent.com/Wynntils/Static-Storage/main/Reference/id_keys.json"

itemdb_path = os.environ.get("ITEMDB_PATH")
shiny_table_path = os.environ.get("SHINY_TABLE_PATH")
id_table_path = os.environ.get("ID_TABLE_PATH")

itemdb_should_update = True
shiny_table_should_update = True
id_table_should_update = True

if itemdb_path:
    itemdb_should_update = False
if shiny_table_path:
    shiny_table_should_update = False
if id_table_path:
    id_table_should_update = False

ITEMDB_PATH = pathlib.Path(itemdb_path or DATA_LOCATION / "itemdb.json")
SHINY_TABLE_PATH = pathlib.Path(shiny_table_path or DATA_LOCATION / "shiny_stats.json")
ID_TABLE_PATH = pathlib.Path(id_table_path or DATA_LOCATION / "id_keys.json")

# Read proxy from environment variable.
http_proxy = os.environ.get("http_proxy") or os.environ.get("HTTP_PROXY")
https_proxy = os.environ.get("https_proxy") or os.environ.get("HTTPS_PROXY")
all_proxy = os.environ.get("all_proxy") or os.environ.get("ALL_PROXY")


@run_async
async def startup():
    """Fetch all data from Wynncraft API and Artemis Repository."""
    client = httpx.AsyncClient(proxy=http_proxy or https_proxy or all_proxy)

    # Fetch the file only if was not updated within 24 hours.
    tasks = []
    if (not ITEMDB_PATH.exists() or time.time() - os.path.getmtime(ITEMDB_PATH) > 86400) and itemdb_should_update:
        tasks += [fetch_item_db(client)]
    if (
        not SHINY_TABLE_PATH.exists() or time.time() - os.path.getmtime(SHINY_TABLE_PATH) > 86400
    ) and shiny_table_should_update:
        tasks += [fetch_shiny_table(client)]
    if (not ID_TABLE_PATH.exists() or time.time() - os.path.getmtime(ID_TABLE_PATH) > 86400) and id_table_should_update:
        tasks += [fetch_id_table(client)]
    await asyncio.gather(*tasks)

    await client.aclose()


async def fetch_item_db(client: httpx.AsyncClient):
    data = await client.get(ITEMDB_FULL_URL)
    with open(DATA_LOCATION / "itemdb.json", "w", encoding="utf-8") as f:
        json.dump(data.json(), f, ensure_ascii=False)


async def fetch_shiny_table(client: httpx.AsyncClient):
    data = await client.get(SHINY_TABLE_URL)
    with open(DATA_LOCATION / "shiny_stats.json", "w", encoding="utf-8") as f:
        json.dump(data.json(), f, ensure_ascii=False)


async def fetch_id_table(client: httpx.AsyncClient):
    data = await client.get(ID_TABLE_URL)
    with open(DATA_LOCATION / "id_keys.json", "w", encoding="utf-8") as f:
        json.dump(data.json(), f, ensure_ascii=False)



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-28 23:09:43
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-02-29 18:18:53
FilePath     : /src/wynntilsresolver/blocks/end.py
"""

from typing import List

from .block import Block


class End(Block):
    _start_byte = 255

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs) -> "End":
        super().from_bytes(data)
        return cls()

    def to_bytes(self) -> List[int]:
        return self.encode_with_start([])

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"End({self._start_byte})"



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-28 23:03:21
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-02-29 22:17:02
FilePath     : /src/wynntilsresolver/blocks/name.py
"""

from typing import List

from .block import Block


class Name(Block):
    _start_byte: int = 2
    name: str

    def __init__(self, name) -> None:
        self.name = name

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs) -> "Name":
        super().from_bytes(data)
        return cls(name=cls.decode_string(data))

    def to_bytes(self) -> List[int]:
        return self.encode_with_start(self.encode_string(self.name))

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Name({self.name})"



"""
Author       : FYWinds i@windis.cn
Date         : 2024-01-13 23:06:58
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-08 15:32:01
FilePath     : /wynntilsresolver/blocks/__init__.py
"""

from .block import Block as Block
from .end import End as End
from .identification import Identification as Identification
from .identification import Identifications as Identifications
from .name import Name as Name
from .powder import Powder as Powder
from .reroll import Reroll as Reroll
from .shiny import Shiny as Shiny
from .type import GearItem as GearItem
from .type import ItemType as ItemType
from .version import Version as Version

__all__ = [
    "Block",
    "Version",
    "ItemType",
    "Name",
    "GearItem",
    "End",
    "Identifications",
    "Reroll",
    "Powder",
    "Shiny",
    "Identification",
]



from typing import Optional


class Resource:
    _instance: Optional["Resource"] = None

    def __init__(self) -> None:
        if Resource._instance is not None:
            raise RuntimeError(f"Multiple initialization of singleton class: {self.__class__.__name__}")
        Resource._instance = self

    @classmethod
    def get(cls) -> "Resource":
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-28 22:32:44
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-02-29 16:20:17
FilePath     : /src/wynntilsresolver/exception.py
"""


class ResolverException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class ParseFailed(ResolverException):
    def __init__(self, message: str):
        super().__init__(message)


class ResolverDefinitionError(ResolverException):
    def __init__(self, message: str):
        super().__init__(message)


class InvalidStartByte(ResolverException):
    def __init__(self, message: str):
        super().__init__(message)


class InvalidItemType(ResolverException):
    def __init__(self, message: str):
        super().__init__(message)


class MissingInfo(ResolverException):
    def __init__(self, message: str):
        super().__init__(message)



"""
Author       : FYWinds i@windis.cn
Date         : 2024-01-13 23:06:52
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-02-29 18:20:02
FilePath     : /src/wynntilsresolver/blocks/version.py
"""

from typing import List

from .block import Block


class Version(Block):
    version: int

    _start_byte: int = 0

    def __init__(self, version) -> None:
        self.version = version

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs) -> "Version":
        super().from_bytes(data)
        v = cls(data[0])
        del data[0]
        return v

    def to_bytes(self) -> List[int]:
        return self.encode_with_start([self.version])

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"Version({self.version})"



"""
Author       : FYWinds i@windis.cn
Date         : 2024-01-13 23:06:55
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-08 15:15:07
FilePath     : /wynntilsresolver/blocks/block.py
"""

from abc import ABC, abstractmethod
from typing import List

from wynntilsresolver.exception import InvalidStartByte


class Block(ABC):
    _start_byte: int

    @classmethod
    def from_bytes(cls, data: List[int], **kwargs):
        # Using deque internally to allow for efficient popping from the left
        if data[0] != cls._start_byte:
            raise InvalidStartByte(f"Invalid start byte for block {cls.__name__}")
        del data[0]

    @abstractmethod
    def to_bytes(self, **kwargs) -> List[int]:
        raise NotImplementedError

    def encode_with_start(self, data: List[int]) -> List[int]:
        return [self._start_byte] + data

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @classmethod
    def decode_string(cls, data: List[int]) -> str:
        name = ""
        for i, byte in enumerate(data):
            if byte == 0:
                break
            name += chr(byte)
        del data[: i + 1]
        return name

    def encode_string(self, string: str) -> List[int]:
        return [ord(x) for x in string] + [0]

    @classmethod
    def decode_variable_sized_int(cls, data: List[int]) -> int:
        """
        Decode an integer using the variable-size encoding scheme.

        The encoding scheme is as follows:
        - The most significant bit is used as a continuation bit
        - The remaining 7 bits are used to store the actual value
        - The number is then decoded using the ZigZag encoding scheme
        """
        num = 0
        shift = 0
        for byte in data:
            num |= (byte & 0x7F) << shift
            if not byte & 0x80:
                break
            shift += 7
        else:
            raise ValueError("Invalid variable-sized integer")
        del data[: (shift // 7) + 1]
        return (num >> 1) ^ -(num & 1)

    def encode_variable_sized_int(self, num: int) -> List[int]:
        """
        Encode an integer using the variable-size encoding scheme.

        The encoding scheme is as follows:
        - The number is encoded using the ZigZag encoding scheme
        - The most significant bit is used as a continuation bit
        - The remaining 7 bits are then used to store the actual value
        """
        # ZigZag encoding
        num = (num << 1) ^ (num >> 0x3F)

        # Encode the integer
        result = []
        while num >= 0x80:
            result.append((num & 0x7F) | 0x80)
            num >>= 7
        result.append(num)
        return result



import asyncio
import functools
import inspect
import sys
import types
from typing import Any, Callable, Coroutine, Dict, List, ParamSpec, TypeVar


def get_annotations_meta(annotation: Dict[str, Any], *, globals=None, locals=None, eval_str=False) -> Dict[str, Any]:
    """Compute the annotations dict for an __annotations__ dict.

    Args:
        annotation (Dict[str, Any]): The __annotations__ dict.
        globals (Dict[str, Any], optional): globals for eval. Defaults to None.
        locals (Dict[str, Any], optional): locals for eval. Defaults to None.
        eval_str (bool, optional): Eval the stringlized type. Defaults to False.

    Returns:
        _type_: The computed annotations dict.
    """
    # get the trigger class of the caller
    frame = inspect.currentframe()
    if frame is None:
        return annotation
    for _ in range(2):
        frame = frame.f_back
        if frame is None:
            return annotation

    module = inspect.getmodule(frame)
    if module is None:
        return annotation
    else:
        obj_globals = globals or getattr(module, "__dict__", None)
        obj_locals = locals or frame.f_locals
        return_value = {
            key: value if not isinstance(value, str) else eval(value, obj_globals, obj_locals)
            for key, value in annotation.items()
        }
        return return_value


def get_annotations(obj, *, globals=None, locals=None, eval_str=False):
    """Compute the annotations dict for an object.
    Copied directly from cpython/Lib/inspect.py#L175

    obj may be a callable, class, or module.
    Passing in an object of any other type raises TypeError.

    Returns a dict.  get_annotations() returns a new dict every time
    it's called; calling it twice on the same object will return two
    different but equivalent dicts.

    This function handles several details for you:

      * If eval_str is true, values of type str will
        be un-stringized using eval().  This is intended
        for use with stringized annotations
        ("from __future__ import annotations").
      * If obj doesn't have an annotations dict, returns an
        empty dict.  (Functions and methods always have anw
        annotations dict; classes, modules, and other types of
        callables may not.)
      * Ignores inherited annotations on classes.  If a class
        doesn't have its own annotations dict, returns an empty dict.
      * All accesses to object members and dict values are done
        using getattr() and dict.get() for safety.
      * Always, always, always returns a freshly-created dict.

    eval_str controls whether or not values of type str are replaced
    with the result of calling eval() on those values:

      * If eval_str is true, eval() is called on values of type str.
      * If eval_str is false (the default), values of type str are unchanged.

    globals and locals are passed in to eval(); see the documentation
    for eval() for more information.  If either globals or locals is
    None, this function may replace that value with a context-specific
    default, contingent on type(obj):

      * If obj is a module, globals defaults to obj.__dict__.
      * If obj is a class, globals defaults to
        sys.modules[obj.__module__].__dict__ and locals
        defaults to the obj class namespace.
      * If obj is a callable, globals defaults to obj.__globals__,
        although if obj is a wrapped function (using
        functools.update_wrapper()) it is first unwrapped.
    """
    if isinstance(obj, type):
        # class
        obj_dict = getattr(obj, "__dict__", None)
        if obj_dict and hasattr(obj_dict, "get"):
            ann = obj_dict.get("__annotations__", None)
            if isinstance(ann, types.GetSetDescriptorType):
                ann = None
        else:
            ann = None

        obj_globals = None
        module_name = getattr(obj, "__module__", None)
        if module_name:
            module = sys.modules.get(module_name, None)
            if module:
                obj_globals = getattr(module, "__dict__", None)
        obj_locals = dict(vars(obj))
        unwrap = obj
    elif isinstance(obj, types.ModuleType):
        # module
        ann = getattr(obj, "__annotations__", None)
        obj_globals = getattr(obj, "__dict__")
        obj_locals = None
        unwrap = None
    elif callable(obj):
        # this includes types.Function, types.BuiltinFunctionType,
        # types.BuiltinMethodType, functools.partial, functools.singledispatch,
        # "class funclike" from Lib/test/test_inspect... on and on it goes.
        ann = getattr(obj, "__annotations__", None)
        obj_globals = getattr(obj, "__globals__", None)
        obj_locals = None
        unwrap = obj
    else:
        raise TypeError(f"{obj!r} is not a module, class, or callable.")

    if ann is None:
        return {}

    if not isinstance(ann, dict):
        raise ValueError(f"{obj!r}.__annotations__ is neither a dict nor None")

    if not ann:
        return {}

    if not eval_str:
        return dict(ann)

    if unwrap is not None:
        while True:
            if hasattr(unwrap, "__wrapped__"):
                unwrap = unwrap.__wrapped__
                continue
            if isinstance(unwrap, functools.partial):
                unwrap = unwrap.func
                continue
            break
        if hasattr(unwrap, "__globals__"):
            obj_globals = unwrap.__globals__

    if globals is None:
        globals = obj_globals
    if locals is None:
        locals = obj_locals

    return_value = {
        key: value if not isinstance(value, str) else eval(value, globals, locals) for key, value in ann.items()
    }
    return return_value


P = ParamSpec("P")
R = TypeVar("R")


def run_async(func: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, R]:
    """A decorator to run a async function synchronously."""

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            return loop.run_until_complete(func(*args, **kwargs))
        else:
            return asyncio.run(func(*args, **kwargs))

    return wrapper


def decode_utf16(data: str) -> List[int]:
    """Decode the int array from utf16 string.

    Args:
        utf16 (str): The utf16 string.

    Returns:
        List[int]: The int array.
    """
    return [int(format(ord(char), "x")[-4:][i : i + 2], 16) for char in data for i in range(0, 4, 2)]
    # li: List[int] = []
    # for char in data:
    #     hex_num = format(ord(char), "04x")
    #     li.append(int(hex_num[-4:-2], 16))
    #     li.append(int(hex_num[-2:], 16))

    # return li


def encode_utf16(data: List[int]) -> str:
    """Encode the utf16 string from int array.

    Args:
        data (List[int]): The int array.

    Returns:
        str: The utf16 string.
    """
    encoded_chars = []

    # remove the possible 0xEE padding when directly converting back
    if data[-1] != 255:
        del data[-1]

    for i in range(0, len(data), 2):
        # more than two byte left
        if i + 1 < len(data):
            # use PUA-A
            char_code = 0xF0000 + (data[i] << 8) + data[i + 1]
        else:
            # use PUA-B and a padding 0xEE
            char_code = 0x100000 + (data[i] << 8) + 0xEE
        encoded_chars.append(chr(char_code))
    return "".join(encoded_chars)



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-28 21:53:42
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-21 13:41:00
FilePath     : /wynntilsresolver/resolver.py
"""

import inspect
from collections import OrderedDict
from dataclasses import field
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    dataclass_transform,
    get_args,
    get_origin,
    overload,
)

from .blocks import Block, End, ItemType, Version
from .exception import InvalidStartByte, ParseFailed, ResolverDefinitionError
from .utils import decode_utf16, encode_utf16, get_annotations_meta


class ResolverMeta(type):
    @classmethod
    def __prepare__(cls, name: str, bases: Tuple[type, ...], /, **kwds: Any):
        return OrderedDict()

    def __new__(cls, name: str, bases: Tuple[type], attrs: Dict[str, Any]):
        annotations: dict[str, type] = get_annotations_meta(attrs.get("__annotations__", {}), eval_str=True)

        # Add _attrs to track all blocks
        if "_attrs" not in attrs:
            attrs["_attrs"] = {}

        if "start" in annotations:
            raise ResolverDefinitionError("start attribute is reserved for Version block")

        # Add start block
        attrs["_attrs"]["start"] = Version
        annotations = {**{"start": Version}, **annotations}

        # type must be provided
        if "item_type" not in annotations or not issubclass(annotations["item_type"], ItemType):
            raise ResolverDefinitionError("Resolver must have a ItemType attribute")

        # Set item_type
        item_type = annotations["item_type"]
        attrs["item_type"] = item_type

        # Add required and optional blocks
        for k, v in annotations.items():
            # If the block is Optional[Block]
            if get_origin(v) == Union and type(None) in (vtype := get_args(v)):
                v = next((x for x in vtype if issubclass(x, Block)), None)
                if v:
                    if v in item_type._required_blocks:
                        raise ResolverDefinitionError(
                            f"Block {k} is in ItemType's required blocks, cannot be optional."
                        )
                    else:
                        if v not in item_type._optional_blocks:
                            item_type._optional_blocks.append(v)
                        attrs["_attrs"][k] = v

            # If the block is Block
            if inspect.isclass(v) and issubclass(v, Block):
                attrs["_attrs"][k] = v

        # Add end block
        if "end" in annotations:
            raise ResolverDefinitionError("end attribute is reserved for end of the resolver")

        attrs["_attrs"]["end"] = End
        annotations = {**annotations, **{"end": End}}

        attrs["__annotations__"] = annotations

        return super().__new__(cls, name, bases, attrs)


T = TypeVar("T", bound="Resolver")


class Resolver(metaclass=ResolverMeta):
    """A wrapper for ResolverMeta."""

    _attrs: Dict[str, Type[Block]]
    item_type: ItemType

    @classmethod
    def from_bytes(cls: Type[T], data: List[int], drop_unknown: bool = False) -> T:
        self = cls()
        data = data.copy()
        attrs = self._attrs.copy()
        required_blocks = {k: v for k, v in attrs.items() if v in self.item_type._required_blocks}
        optional_blocks = {k: v for k, v in attrs.items() if v in self.item_type._optional_blocks}
        parsed_blocks = []

        while data:
            k = None
            try:
                for k, v in attrs.items():
                    if data[0] == v._start_byte:
                        required_blocks.pop(k, None)
                        optional_blocks.pop(k, None)
                        setattr(self, k, attrs.pop(k).from_bytes(data=data, parsed_blocks=parsed_blocks))

                        parsed_blocks.append(getattr(self, k))

                        # ignore the padding 0xEE or other unknown bytes
                        if v == End:
                            data = []
                        break
                else:
                    if drop_unknown:
                        del data[0]
                    else:
                        raise InvalidStartByte(f"Unknown registered start byte {data[0]}")
            except Exception as e:
                raise ParseFailed(f"Failed to parse. Error during parsing block: {k}") from e

        # Ended with unparsed required blocks
        if required_blocks:
            raise ParseFailed(f"Failed to parse. Missing blocks: {required_blocks}")

        # Set all unparsed optional block to None
        for k, v in optional_blocks.items():
            setattr(self, k, None)

        return self

    def to_bytes(self) -> List[int]:
        data = []
        for k, v in self._attrs.items():
            if (block := getattr(self, k)) is not None:
                data.extend(block.to_bytes())
        return data

    @classmethod
    def from_utf16(cls: Type[T], data: str, drop_unknown: bool = False) -> T:
        return cls.from_bytes(decode_utf16(data), drop_unknown)

    def to_utf16(self):
        return encode_utf16(self.to_bytes())

    def __repr__(self) -> str:
        repr = f"<class '{self.__class__.__name__}' {{"
        for k, v in self._attrs.items():
            repr += f"{k}={getattr(self, k).__repr__()}, "

        repr = repr[:-2] + "} >"
        return repr

    def __str__(self) -> str:
        repr = f"<class '{self.__class__.__name__}' {{"
        for k, v in self._attrs.items():
            repr += f"{k}={getattr(self, k)}, "

        repr = repr[:-2] + "} >"
        return repr



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-29 13:23:08
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-25 02:39:22
FilePath     : /wynntilsresolver/blocks/identification.py
"""

import json
import math
from dataclasses import dataclass
from typing import Dict, List

from wynntilsresolver.exception import MissingInfo
from wynntilsresolver.startup import ID_TABLE_PATH, ITEMDB_PATH

from .block import Block
from .name import Name

with open(ITEMDB_PATH, encoding="utf-8") as f:
    ITEMDB = json.load(f)

with open(ID_TABLE_PATH, encoding="utf-8") as f:
    ID_TABLE = json.load(f)


def _extract_item_name(blocks: List[Block]) -> str:
    for block in blocks:
        if isinstance(block, Name):
            return block.name
    else:
        raise MissingInfo("Item name not found when trying to parse identifications.")


def _id_from_str(id: str) -> int:
    return ID_TABLE[id]


def _id_from_int(id: int) -> str:
    return next(k for k, v in ID_TABLE.items() if v == id)


@dataclass
class Identification:
    id: str
    """Key of the identification. From Wynncraft official API."""
    internal_id: int
    """Numeric ID of the identification. From Artemis Data."""
    base: int
    """Base value of the identification. same as value if pre-identified."""
    roll: int
    """Roll of the identification. -1 represents pre-identified."""
    value: int
    """Value of the identification. Two digits after the decimal point are povided if comes from roll."""

    @classmethod
    def from_simple(cls, id: str, internal_id: int, meta: Dict[str, int], roll: int) -> "Identification":
        id_base = meta["raw"]
        id_value = cls.special_round(id_base * (roll / 100))
        return cls(id, internal_id, id_base, roll, id_value)

    @classmethod
    def from_extend(cls, id, data: List[int]) -> "Identification":
        base = Block.decode_variable_sized_int(data)
        roll = data[0]
        del data[0]
        id_name = _id_from_int(id)
        id_value = cls.special_round(base * (roll / 100))
        return cls(id_name, id, base, roll, id_value)

    def to_simple(self) -> List[int]:
        # TODO
        raise NotImplementedError

    def to_extend(self) -> List[int]:
        # TODO
        raise NotImplementedError

    @staticmethod
    def special_round(num: float) -> int:
        return math.floor(num + 0.5)


class Identifications(Block):
    _start_byte = 3
    identifications: List[Identification]

    def __init__(self, identifications: List[Identification]) -> None:
        self.identifications = identifications

    @classmethod
    def from_bytes(cls, data: List[int], parsed_blocks: List[Block], **kwargs) -> "Identifications":
        super().from_bytes(data)
        # number of non-pre-identified ids
        id_num = data[0]
        del data[0]
        # 1 -> extended encoding
        extend = bool(data[0])
        del data[0]

        identifications: List[Identification] = []
        name = _extract_item_name(parsed_blocks)
        item_identifications_meta = ITEMDB[name]["identifications"]

        if not extend:
            for id, meta in item_identifications_meta.items():
                if isinstance(meta, int):
                    # pre-identified
                    identifications.append(Identification(id, _id_from_str(id), meta, -1, meta))

            # Truncate data to the length of ids
            id_data = data[: id_num * 2]
            del data[: id_num * 2]

            for id, roll in zip(id_data[::2], id_data[1::2]):
                id_name = _id_from_int(id)
                id_meta = item_identifications_meta[id_name]
                identifications.append(Identification.from_simple(id_name, id, id_meta, roll))

        else:
            # Number of pre-identified ids
            id_num_pre = data[0]
            del data[0]
            for _ in range(id_num_pre):
                id = data[0]
                del data[0]
                id_str = _id_from_int(id)
                value = cls.decode_variable_sized_int(data)
                identifications.append(Identification(id_str, id, value, -1, value))
            for _ in range(id_num):
                id = data[0]
                del data[0]
                identifications.append(Identification.from_extend(id, data))

        return cls(identifications)

    def to_bytes(self, parsed_blocks: List[Block], extend: bool = False) -> List[int]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"Identifications({self.identifications})"

    def __str__(self) -> str:
        return f"{self.identifications}"



"""
Author       : FYWinds i@windis.cn
Date         : 2023-12-14 10:47:00
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-21 13:01:18
FilePath     : /wynntilsresolver/__init__.py
"""

from .startup import startup

startup()

from .item import GearItemResolver as GearItemResolver
from .resolver import Resolver as Resolver

__all__ = ["Resolver", "GearItemResolver"]



"""
Author       : FYWinds i@windis.cn
Date         : 2024-03-01 16:01:35
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-03-01 20:08:24
FilePath     : /src/wynntilsresolver/blocks/shiny.py
"""

import json
from typing import Dict, List

from wynntilsresolver.startup import SHINY_TABLE_PATH

from .block import Block

with open(SHINY_TABLE_PATH, encoding="utf-8") as f:
    shiny_table: List[Dict] = json.load(f)


class Shiny(Block):
    _start_byte: int = 6

    name: str
    """Key of the shiny. From Artemis Data."""
    internal_id: int
    """Internal ID of the shiny. From Artemis Data."""
    display_name: str
    """Display name of the shiny. From Artemis Data."""
    value: int
    """Value of the shiny."""
    reroll: int = 0
    """Reroll count of the shiny."""

    def __init__(self, name: str, internal_id: int, display_name: str, value: int, reroll: int) -> None:
        self.name = name
        self.internal_id = internal_id
        self.display_name = display_name
        self.value = value
        self.reroll = reroll

    @classmethod
    def from_bytes(cls, data, **kwargs) -> "Shiny":
        super().from_bytes(data)
        internal_id = data[0]
        del data[0]
        reroll = data[0] if data else 0
        del data[0]
        value = cls.decode_variable_sized_int(data)
        for shiny in shiny_table:
            if shiny["id"] == internal_id:
                return cls(shiny["key"], internal_id, shiny["displayName"], value, reroll)

        raise ValueError(f"Shiny with internal ID {internal_id} not found in shiny table.")

    def to_bytes(self, **kwargs) -> List[int]:
        return self.encode_with_start([self.internal_id] + self.encode_variable_sized_int(self.value))

    def __str__(self) -> str:
        return f"Shiny(name={self.name}, internal_id={self.internal_id}, display_name={self.display_name}, value={self.value}, reroll={self.reroll})"

    def __repr__(self) -> str:
        return f"Shiny(name={self.name}, internal_id={self.internal_id}, display_name={self.display_name}, value={self.value}, reroll={self.reroll})"



"""
Author       : FYWinds i@windis.cn
Date         : 2024-02-29 18:50:29
LastEditors  : FYWinds i@windis.cn
LastEditTime : 2024-02-29 22:43:35
FilePath     : /src/wynntilsresolver/blocks/reroll.py
"""

from .block import Block


class Reroll(Block):
    _start_byte = 5
    rerolls: int

    def __init__(self, rerolls: int) -> None:
        self.rerolls = rerolls

    @classmethod
    def from_bytes(cls, data, **kwargs) -> "Reroll":
        super().from_bytes(data)
        reroll = cls(data[0])
        del data[0]
        return reroll

    def to_bytes(self):
        return self.encode_with_start([self.rerolls])

    def __str__(self) -> str:
        return f"{self.rerolls}"

    def __repr__(self) -> str:
        return f"Reroll({self.rerolls})"
